@@ -303,8 +303,11 @@ sub _generate_request_handler {
         my $delay  = shift;
         my $data   = shift;
         my $status = shift || 200;
-        my $format = $config->{responses}{$status} || $config->{responses}{default} || {};
-        my @err    = $self->_validator->validate($data, $format->{schema});
+        my $format = $config->{responses}{$status} || $config->{responses}{default} || undef;
+        my @err
+          = !$format          ? $self->_validator->validate($data, undef)
+          : $format->{schema} ? $self->_validator->validate($data, $format->{schema})
+          :                     ();
 
         return $c->render_swagger({errors => \@err, valid => Mojo::JSON->false}, $data, 500) if @err;
         return $c->render_swagger({}, $data, $status);
@@ -326,6 +329,7 @@ sub _not_implemented {
 
 sub _validate_input {
   my ($self, $c, $config) = @_;
+  my $body    = $c->req->body_params;
   my $headers = $c->req->headers;
   my $query   = $c->req->url->query;
   my (%input, %v);
@@ -339,7 +343,7 @@ sub _validate_input {
       : $in eq 'path'     ? $c->stash($name)
       : $in eq 'header'   ? $headers->header($name)
       : $in eq 'body'     ? $c->req->json
-      : $in eq 'formData' ? $c->req->body_params->to_hash
+      : $in eq 'formData' ? $body->param($name)
       :                     "Invalid 'in' for parameter $name in schema definition";
 
     $value //= $p->{default};
@@ -349,16 +353,20 @@ sub _validate_input {
       $value += 0 if $type =~ /^(?:integer|number)/ and $value =~ /^-?\d/;
       $value = ($value eq 'false' or !$value) ? Mojo::JSON->false : Mojo::JSON->true if $type eq 'boolean';
 
-      if ($in eq 'body' or $in eq 'formData') {
+      if ($in eq 'body') {
         warn "[Swagger2] Validate $in @{[$c->req->body]}\n" if DEBUG;
         push @e,
           map { $_->{path} = $_->{path} eq "/" ? "/$name" : "/$name$_->{path}"; $_; }
           $self->_validator->validate($value, $p->{schema});
       }
-      else {
+      elsif (defined $value) {
         warn "[Swagger2] Validate $in $name=$value\n" if DEBUG;
         push @e, $self->_validator->validate({$name => $value}, {properties => {$name => $p}});
       }
+      else {
+        warn "[Swagger2] Validate $in $name=undef()\n" if DEBUG;
+        push @e, $self->_validator->validate({}, {properties => {$name => $p}});
+      }
     }
 
     $input{$name} = $value unless @e;
